from shapely.geometry.polygon import Polygon
import numpy as np

from ..spawner.poly import sprinkle_cube, simple_cube
from .empty_room import EmptyRoomConfig, EmptyRoomGenerator

class ObstacleRoomConfig(EmptyRoomConfig):
    def __init__(self, wall_shape, obstacle_shape, obstacle_count):
        super(ObstacleRoomConfig, self).__init__(wall_shape)

        self.obstacle_shape = obstacle_shape
        self.obstacle_count = obstacle_count

        self.tag_wall = 'wall'
        self.tag_obstacle = 'obstacle'

    def prepare(self):
        self.register(self.tag_wall, self.wall_shape, 1)
        self.register(self.tag_obstacle, self.obstacle_shape, self.obstacle_count)

class ObstacleRoomGenerator(EmptyRoomGenerator):
    def __init__(self, 
                 obstacle_count=10,
                 obstacle_size=0.7,
                 room_length_max=9,
                 room_wall_thickness=0.05,
                 wall_threshold=0.1):
        
        super(ObstacleRoomGenerator, self).__init__(room_length_max, room_wall_thickness)
        self.wall_threshold = wall_threshold
        self.obstacle_count = obstacle_count
        self.obstacle_size = obstacle_size

    def generate_new(self):
        wall_shape = self._create_wall_poly()
        wall_collision = True
        wall_pos = np.array([(0,0,0)])

        wall_interior = Polygon(wall_shape.interiors[0])
        xy, yaw, polys = sprinkle_cube(
            area_poly=wall_interior, 
            count=self.obstacle_count, 
            cube_size=self.obstacle_size,
            interior_thresh=self.wall_threshold
        )
        obstacle_shape = simple_cube(self.obstacle_size)
        obstacle_collision = True
        obstacle_pos = np.empty([len(xy), 3])
        obstacle_pos[:,:2] = xy
        obstacle_pos[:,2] = yaw

        obs_room = ObstacleRoomConfig(
            wall_shape=wall_shape,
            obstacle_shape=obstacle_shape,
            obstacle_count=self.obstacle_count
        )
        obs_room.prepare()
        obs_room.set_config(obs_room.tag_wall, wall_collision, wall_pos)
        obs_room.set_config(obs_room.tag_obstacle, obstacle_collision, obstacle_pos)
        obs_room.set_polygons_auto(obs_room.tag_wall)
        obs_room.set_polygons_direct(obs_room.tag_obstacle, polys)

        return obs_room



import numpy as np
import quaternion

def vec_to_transform_matrix(vec):
    a = np.identity(len(vec)+1)
    a[:len(vec),len(vec)] = vec
    return a

def radian_to_rotation_matrix(radian):
    return np.array([
        [np.cos(radian), -np.sin(radian), 0],
        [np.sin(radian), np.cos(radian), 0],
        [0,0,1]
    ])

def get_square_horizon(base_xy, radius, z_angle=0):
    d1 = [1,1,-1,-1]
    d2 = [1,-1,-1,1]
    poss = np.array([np.quaternion(1,d1[i]*radius,d2[i]*radius,0) for i in range(4)])
    q = quaternion.from_euler_angles(0,0,z_angle)
    pq = q*poss*q.conj()
    return np.array([[p.x, p.y] for p in pq])+base_xy

def add_dimension(vecs, value=0):
    return np.concatenate([vecs[np.newaxis,:,i].T if i<len(vecs[0]) else np.full([len(vecs),1], value) for i in range(len(vecs[0])+1)], axis=1)

def convert_180(rad360):
    '''
    [0,360] -> [-180,180]
    '''
    return ((rad360 - np.pi) % (np.pi*2)) - np.pi

def convert_360(rad180):
    '''
    [-180,180] -> [0,360]
    '''
    return rad180 % (np.pi*2)

def transform_2d(target_x, target_y, target_yaw, origin_x, origin_y, origin_yaw):
    t_mat = np.identity(3)
    t_mat[:2,2] = [-origin_x, -origin_y]
    r_mat = np.array([
                [np.cos(-origin_yaw), -np.sin(-origin_yaw), 0],
                [np.sin(-origin_yaw), np.cos(-origin_yaw), 0],
                [0, 0, 1]
            ])
    af = np.dot(r_mat, t_mat)
    target_xy = np.array([target_x, target_y, 1])
    transed_xy = np.dot(af, target_xy)
    
    return transed_xy[0], transed_xy[1], convert_180(target_yaw - origin_yaw)

def get_affine(x, y, yaw):
    """
    Notice: Caluclation with right-handed coordinate system.
    """
    rotate = radian_to_rotation_matrix(yaw)
    # trans = vec_to_transform_matrix([y,-x])
    trans = vec_to_transform_matrix([x,y])
    affine = np.dot(trans, rotate)
    return affine



import numpy as np
from shapely.affinity import affine_transform

from shapely.geometry import Polygon, MultiPoint, MultiPolygon
from shapely.ops import unary_union, triangulate

from ..utils import get_square_horizon, get_affine
from . import geom

def simple_cube(cube_size, xy=(0,0), yaw=0):
    return Polygon(get_square_horizon(xy, cube_size/2, yaw))

def sprinkle_cube(area_poly, count, cube_size, interior_thresh=0):
    xy, yaw = geom.sample_sprinkle(area_poly, count, cube_size+interior_thresh)
    return xy, yaw, [Polygon(get_square_horizon(p, cube_size/2, r)) for p, r in zip(xy, yaw)]

def create_zones(polys, label):
    parray = np.array(polys)
    cls_count = max(label)+1

    zone_polys = [None]*cls_count
    zone_hulls = [None]*cls_count

    for i in range(cls_count):
        cluster = parray[label==i]
        vertices = [p.exterior.coords for p in cluster]
        vertices_all = np.concatenate(vertices)
        zone_polys[i] = unary_union(cluster)
        zone_hulls[i] = MultiPoint(vertices_all).convex_hull
    
    return zone_polys, zone_hulls

def get_clustered_zones(polys, thresh):
    label, _ = geom.get_cluster(polys, thresh)
    return create_zones(polys, label)

def random_triangulation(n_points=10, x_min=-10, x_max=10, y_min=-10, y_max=10):
    assert x_min < x_max, 'x_min must be lower than x_max'
    assert y_min < y_max, 'y_min must be lower than y_max'
    points = np.random.random([n_points, 2])
    points[:,0] = points[:,0] * (x_max - x_min) + x_min
    points[:,1] = points[:,1] * (y_max - y_min) + y_min
    return unary_union(triangulate(MultiPoint(points)))

def get_moved_poly(base_geom, x, y, yaw):
    aff = get_affine(x,y,yaw)
    return affine_transform(base_geom, [aff[0,0], aff[0,1], aff[1,0], aff[1,1], aff[0,2], aff[1,2]])


from shapely.geometry.polygon import Polygon

from ..spawner.poly import random_triangulation
from ..generator_factory import RoomConfig, RoomGeneratorFactory

class EmptyRoomConfig(RoomConfig):
    def __init__(self, wall_shape):
        super(EmptyRoomConfig, self).__init__()
        self.wall_shape = wall_shape

        self.wall_interior_polygon = Polygon(self.wall_shape.interiors[0])
        self.wall_exterior_polygon = None

        self.tag_wall = 'wall'

    def prepare(self):
        self.register(self.tag_wall, self.wall_shape, 1)

    def get_freespace_poly(self):
        return self.get_space_poly(self.tag_wall)

class EmptyRoomGenerator(RoomGeneratorFactory):
    def __init__(self, 
                 room_length_max=9,
                 room_wall_thickness=0.05):
        
        super(EmptyRoomGenerator, self).__init__()
        self.room_length_max = room_length_max
        self.room_wall_thickness = room_wall_thickness

    def _create_wall_poly(self):
        n_points = self.room_length_max * 6
        x_max = self.room_length_max / 2
        x_min = -x_max
        y_max = x_max
        y_min = -x_max
        p = random_triangulation(n_points, x_min, x_max, y_min, y_max)
        return p.exterior.buffer(self.room_wall_thickness, join_style=2) # mitre style

    def generate_new(self):
        wall_shape = self._create_wall_poly()
        wall_collision = True
        wall_pos = [(0,0,0)]

        emp_room = EmptyRoomConfig(
            wall_shape=wall_shape
        )

        emp_room.prepare()
        emp_room.set_config(emp_room.tag_wall, wall_collision, wall_pos)
        emp_room.set_polygons_auto(emp_room.tag_wall)

        return emp_room


from shapely.geometry.polygon import Polygon
from shapely.ops import unary_union
import numpy as np

from ..spawner.poly import sprinkle_cube, simple_cube, get_clustered_zones
from ..spawner.geom import sample_from_faces
from .obstacle_room import ObstacleRoomConfig, ObstacleRoomGenerator

class SimpleSearchRoomConfig(ObstacleRoomConfig):
    def __init__(self, 
                 wall_shape, 
                 obstacle_shape, 
                 target_shape, 
                 obstacle_count, 
                 target_count, 
                 obstacle_hulls):

        super(SimpleSearchRoomConfig, self).__init__(wall_shape, obstacle_shape, obstacle_count)

        self.target_shape = target_shape
        self.target_count = target_count
        self.obstacle_hulls = obstacle_hulls

        self.tag_wall = 'wall'
        self.tag_obstacle = 'obstacle'
        self.tag_target = 'target'

    def prepare(self):
        self.register(self.tag_wall, self.wall_shape, 1)
        self.register(self.tag_obstacle, self.obstacle_shape, self.obstacle_count)
        self.register(self.tag_target, self.target_shape, self.target_count)

    def get_freezone_poly(self):
        return self.get_inner_poly(self.tag_wall, unary_union(self.obstacle_hulls))

class SimpleSearchRoomGenerator(ObstacleRoomGenerator):
    target_each_count = 1
    target_sample_face = 0.01

    def __init__(self, 
                 obstacle_count=10,
                 obstacle_size=0.7,
                 target_size=0.2,
                 obstacle_zone_thresh=1.5,
                 room_length_max=9,
                 room_wall_thickness=0.05,
                 wall_threshold=0.1):

        super(SimpleSearchRoomGenerator, self).__init__(obstacle_count, obstacle_size, room_length_max, room_wall_thickness, wall_threshold)

        self.obstacle_zone_thresh = obstacle_zone_thresh
        self.target_size = target_size

    def generate_new(self):
        ## ------------------------- same super's ------------------------------------
        wall_shape = self._create_wall_poly()
        wall_collision = True
        wall_pos = np.array([(0,0,0)])

        wall_interior = Polygon(wall_shape.interiors[0])
        obstacle_xy, obstacle_yaw, obstacle_polys = sprinkle_cube(
            area_poly=wall_interior, 
            count=self.obstacle_count, 
            cube_size=self.obstacle_size,
            interior_thresh=self.wall_threshold
        )
        obstacle_shape = simple_cube(self.obstacle_size)
        obstacle_collision = True
        obstacle_pos = np.empty([len(obstacle_xy), 3])
        obstacle_pos[:,:2] = obstacle_xy
        obstacle_pos[:,2] = obstacle_yaw
        ## ---------------------------------------------------------------------------

        zone_polys, zone_hull = get_clustered_zones(obstacle_polys, self.obstacle_zone_thresh)
        target_shape = simple_cube(self.target_size)
        target_collision = False
        target_pos = self._sample_target_pos(zone_hull)

        room = SimpleSearchRoomConfig(
            wall_shape=wall_shape,
            obstacle_shape=obstacle_shape,
            target_shape=target_shape,
            obstacle_count=self.obstacle_count,
            target_count=len(target_pos),
            obstacle_hulls=zone_hull
        )

        room.prepare()
        room.set_config(room.tag_wall, wall_collision, wall_pos)
        room.set_config(room.tag_obstacle, obstacle_collision, obstacle_pos)
        room.set_config(room.tag_target, target_collision, target_pos)
        room.set_polygons_auto(room.tag_wall)
        room.set_polygons_direct(room.tag_obstacle, obstacle_polys)
        room.set_polygons_auto(room.tag_target)

        return room

    def _sample_target_pos(self, hulls):
        target_placing_hull = [h.buffer(self.wall_threshold) for h in hulls]
        target_pos = np.empty([len(hulls), 3])
        target_pos[:,:2] = sample_from_faces(
            polys=target_placing_hull, 
            count=self.target_each_count, 
            face_size=self.target_sample_face
        )[:,0]
        target_pos[:,2] = 0.0
        return target_pos
    
    def reposition_target(self, room_conf):
        target_pos = self._sample_target_pos(room_conf.obstacle_hulls)
        room_conf.set_config_positions(room_conf.tag_target, target_pos)
        room_conf.set_polygons_auto(room_conf.tag_target)


from shapely.geometry.polygon import Polygon
from shapely.ops import unary_union
import numpy as np

from randoor.spawner.poly import simple_cube

from ..spawner.geom import sample_sure
from .simple_search_room import SimpleSearchRoomGenerator, SimpleSearchRoomConfig

class ChestSearchRoomConfig(SimpleSearchRoomConfig):
    def __init__(self, 
                 wall_shape, 
                 obstacle_shape, 
                 target_shape, 
                 key_shape, 
                 obstacle_count, 
                 target_count, 
                 key_count, 
                 obstacle_hulls, 
                 key_placing_area):

        super(ChestSearchRoomConfig, self).__init__(wall_shape, obstacle_shape, target_shape, obstacle_count, target_count, obstacle_hulls)

        self.key_shape = key_shape
        self.key_count = key_count
        self.key_placing_area = key_placing_area

        self.tag_wall = 'wall'
        self.tag_obstacle = 'obstacle'
        self.tag_target = 'target'
        self.tag_key = 'key'

    def prepare(self):
        self.register(self.tag_wall, self.wall_shape, 1)
        self.register(self.tag_obstacle, self.obstacle_shape, self.obstacle_count)
        self.register(self.tag_target, self.target_shape, self.target_count)
        self.register(self.tag_key, self.key_shape, self.key_count)

    def get_freezone_poly(self):
        return self.get_inner_poly(self.tag_wall, unary_union(self.obstacle_hulls))

class ChestSearchRoomGenerator(SimpleSearchRoomGenerator):
    key_each_count = 1

    def __init__(self, 
                 obstacle_count=10,
                 obstacle_size=0.7,
                 target_size=0.2,
                 key_size=0.2,
                 obstacle_zone_thresh=1.5,
                 distance_key_placing=0.7,
                 range_key_placing=0.3,
                 room_length_max=9,
                 room_wall_thickness=0.05,
                 wall_threshold=0.1):

        super(ChestSearchRoomGenerator, self).__init__(obstacle_count, obstacle_size, target_size, obstacle_zone_thresh, room_length_max, room_wall_thickness, wall_threshold)

        self.key_size = key_size
        self.distance_key_placing = distance_key_placing
        self.range_key_placing = range_key_placing

    def generate_new(self):
        pre =  super(ChestSearchRoomGenerator, self).generate_new()
        freezone = pre.get_freezone_poly()
        zone_hull = pre.obstacle_hulls
        hull_buff = self.distance_key_placing + self.range_key_placing
        path_area = freezone.buffer(-self.distance_key_placing)
        key_placing_area = [path_area.intersection(h.buffer(hull_buff)) for h in zone_hull]

        key_shape = simple_cube(self.key_size)
        key_collision = False
        key_pos = self._sample_key_pos(key_placing_area)

        room = ChestSearchRoomConfig(
            wall_shape=pre.wall_shape,
            obstacle_shape=pre.obstacle_shape,
            target_shape=pre.target_shape,
            key_shape=key_shape,
            obstacle_count=pre.obstacle_count,
            target_count=pre.target_count,
            key_count=len(key_pos),
            obstacle_hulls=pre.obstacle_hulls,
            key_placing_area=key_placing_area
        )

        room.prepare()
        self.merge_config(pre, room)
        room.set_config(room.tag_key, key_collision, key_pos)
        room.set_polygons_auto(room.tag_key)

        return room

    def _sample_key_pos(self, key_placing_areas):
        key_pos = np.empty([len(key_placing_areas), 3])
        key_pos[:,:2] = [sample_sure(a, self.key_each_count, self.range_key_placing)[0] for a in key_placing_areas]
        key_pos[:,2] = 0.0
        return key_pos

    def reposition_key(self, room_conf):
        key_pos = self._sample_key_pos(room_conf.key_placing_area)
        room_conf.set_config_positions(room_conf.tag_key, key_pos)
        room_conf.set_polygons_auto(room_conf.tag_key)